@@ -3,15 +3,22 @@ This module tests the functionality of shapefile.py.
 """
 # std lib imports
 import os.path
+import sys
+if sys.version_info.major == 3:
+    from pathlib import Path
 
 # third party imports
 import pytest
 import json
 import datetime
+if sys.version_info.major == 2:
+    # required by pytest for python <36
+    from pathlib2 import Path
 
 # our imports
 import shapefile
 
+
 # define various test shape tuples of (type, points, parts indexes, and expected geo interface output)
 geo_interface_tests = [ (shapefile.POINT, # point
                             [(1,1)],
@@ -42,7 +49,7 @@ geo_interface_tests = [ (shapefile.POINT, # point
                              ],
                             [0],
                             {'type':'Polygon','coordinates':[
-                                shapefile.rewind([(1,1),(1,9),(9,9),(9,1),(1,1)]),
+                                [(1,1),(1,9),(9,9),(9,1),(1,1)],
                                 ]}
                         ),
                        (shapefile.POLYGON, # single polygon, holes (ordered)
@@ -52,9 +59,9 @@ geo_interface_tests = [ (shapefile.POINT, # point
                              ],
                             [0,5,5+5],
                             {'type':'Polygon','coordinates':[
-                                shapefile.rewind([(1,1),(1,9),(9,9),(9,1),(1,1)]), # exterior
-                                shapefile.rewind([(2,2),(4,2),(4,4),(2,4),(2,2)]), # hole 1
-                                shapefile.rewind([(5,5),(7,5),(7,7),(5,7),(5,5)]), # hole 2
+                                [(1,1),(1,9),(9,9),(9,1),(1,1)], # exterior
+                                [(2,2),(4,2),(4,4),(2,4),(2,2)], # hole 1
+                                [(5,5),(7,5),(7,7),(5,7),(5,5)], # hole 2
                                 ]}
                         ),
                        (shapefile.POLYGON, # single polygon, holes (unordered)
@@ -65,9 +72,9 @@ geo_interface_tests = [ (shapefile.POINT, # point
                              ],
                             [0,5,5+5],
                             {'type':'Polygon','coordinates':[
-                                shapefile.rewind([(1,1),(1,9),(9,9),(9,1),(1,1)]), # exterior
-                                shapefile.rewind([(2,2),(4,2),(4,4),(2,4),(2,2)]), # hole 1
-                                shapefile.rewind([(5,5),(7,5),(7,7),(5,7),(5,5)]), # hole 2
+                                [(1,1),(1,9),(9,9),(9,1),(1,1)], # exterior
+                                [(2,2),(4,2),(4,4),(2,4),(2,2)], # hole 1
+                                [(5,5),(7,5),(7,7),(5,7),(5,5)], # hole 2
                                 ]}
                         ),
                        (shapefile.POLYGON, # multi polygon, no holes
@@ -77,10 +84,10 @@ geo_interface_tests = [ (shapefile.POINT, # point
                             [0,5],
                             {'type':'MultiPolygon','coordinates':[
                                 [ # poly 1
-                                    shapefile.rewind([(1,1),(1,9),(9,9),(9,1),(1,1)]),
+                                    [(1,1),(1,9),(9,9),(9,1),(1,1)],
                                 ],
                                 [ # poly 2
-                                    shapefile.rewind([(11,11),(11,19),(19,19),(19,11),(11,11)]),
+                                    [(11,11),(11,19),(19,19),(19,11),(11,11)],
                                 ],
                                 ]}
                         ),
@@ -95,14 +102,14 @@ geo_interface_tests = [ (shapefile.POINT, # point
                             [0,5,10,15,20,25],
                             {'type':'MultiPolygon','coordinates':[
                                 [ # poly 1
-                                    shapefile.rewind([(1,1),(1,9),(9,9),(9,1),(1,1)]), # exterior
-                                    shapefile.rewind([(2,2),(4,2),(4,4),(2,4),(2,2)]), # hole 1
-                                    shapefile.rewind([(5,5),(7,5),(7,7),(5,7),(5,5)]), # hole 2
+                                    [(1,1),(1,9),(9,9),(9,1),(1,1)], # exterior
+                                    [(2,2),(4,2),(4,4),(2,4),(2,2)], # hole 1
+                                    [(5,5),(7,5),(7,7),(5,7),(5,5)], # hole 2
                                 ],
                                 [ # poly 2
-                                    shapefile.rewind([(11,11),(11,19),(19,19),(19,11),(11,11)]), # exterior
-                                    shapefile.rewind([(12,12),(14,12),(14,14),(12,14),(12,12)]), # hole 1
-                                    shapefile.rewind([(15,15),(17,15),(17,17),(15,17),(15,15)]), # hole 2
+                                    [(11,11),(11,19),(19,19),(19,11),(11,11)], # exterior
+                                    [(12,12),(14,12),(14,14),(12,14),(12,12)], # hole 1
+                                    [(15,15),(17,15),(17,17),(15,17),(15,15)], # hole 2
                                 ],
                                 ]}
                         ),
@@ -116,15 +123,15 @@ geo_interface_tests = [ (shapefile.POINT, # point
                             [0,5,10,15,20],
                             {'type':'MultiPolygon','coordinates':[
                                 [ # poly 1
-                                    shapefile.rewind([(1,1),(1,9),(9,9),(9,1),(1,1)]), # exterior 1
-                                    shapefile.rewind([(2,2),(8,2),(8,8),(2,8),(2,2)]), # hole 1.1
+                                    [(1,1),(1,9),(9,9),(9,1),(1,1)], # exterior 1
+                                    [(2,2),(8,2),(8,8),(2,8),(2,2)], # hole 1.1
                                 ],
                                 [ # poly 2
-                                    shapefile.rewind([(3,3),(3,7),(7,7),(7,3),(3,3)]), # exterior 2
-                                    shapefile.rewind([(4,4),(6,4),(6,6),(4,6),(4,4)]), # hole 2.1
+                                    [(3,3),(3,7),(7,7),(7,3),(3,3)], # exterior 2
+                                    [(4,4),(6,4),(6,6),(4,6),(4,4)], # hole 2.1
                                 ],
                                 [ # poly 3
-                                    shapefile.rewind([(4.5,4.5),(4.5,5.5),(5.5,5.5),(5.5,4.5),(4.5,4.5)]), # exterior 3
+                                    [(4.5,4.5),(4.5,5.5),(5.5,5.5),(5.5,4.5),(4.5,4.5)], # exterior 3
                                 ],
                                 ]}
                         ),
@@ -138,15 +145,15 @@ geo_interface_tests = [ (shapefile.POINT, # point
                             [0,5,10,15,20+3],
                             {'type':'MultiPolygon','coordinates':[
                                 [ # poly 1
-                                    shapefile.rewind([(1,1),(1,9),(9,9),(9,1),(1,1)]), # exterior 1
-                                    shapefile.rewind([(2,2),(3,3),(4,2),(8,2),(8,8),(4,8),(2,8),(2,4),(2,2)]), # hole 1.1
+                                    [(1,1),(1,9),(9,9),(9,1),(1,1)], # exterior 1
+                                    [(2,2),(3,3),(4,2),(8,2),(8,8),(4,8),(2,8),(2,4),(2,2)], # hole 1.1
                                 ],
                                 [ # poly 2
-                                    shapefile.rewind([(3,3),(3,7),(7,7),(7,3),(3,3)]), # exterior 2
-                                    shapefile.rewind([(4,4),(4,4),(6,4),(6,4),(6,4),(6,6),(4,6),(4,4)]), # hole 2.1
+                                    [(3,3),(3,7),(7,7),(7,3),(3,3)], # exterior 2
+                                    [(4,4),(4,4),(6,4),(6,4),(6,4),(6,6),(4,6),(4,4)], # hole 2.1
                                 ],
                                 [ # poly 3
-                                    shapefile.rewind([(4.5,4.5),(4.5,5.5),(5.5,5.5),(5.5,4.5),(4.5,4.5)]), # exterior 3
+                                    [(4.5,4.5),(4.5,5.5),(5.5,5.5),(5.5,4.5),(4.5,4.5)], # exterior 3
                                 ],
                                 ]}
                         ),
@@ -162,17 +169,16 @@ geo_interface_tests = [ (shapefile.POINT, # point
                             [0,5,10,15,20,25,30],
                             {'type':'MultiPolygon','coordinates':[
                                 [ # poly 1
-                                    shapefile.rewind([(1,1),(1,9),(9,9),(9,1),(1,1)]), # exterior
-                                    shapefile.rewind([(2,2),(4,2),(4,4),(2,4),(2,2)]), # hole 1
-                                    shapefile.rewind([(5,5),(7,5),(7,7),(5,7),(5,5)]), # hole 2
+                                    [(1,1),(1,9),(9,9),(9,1),(1,1)], # exterior
+                                    [(2,2),(4,2),(4,4),(2,4),(2,2)], # hole 1
+                                    [(5,5),(7,5),(7,7),(5,7),(5,5)], # hole 2
                                 ],
                                 [ # poly 2
-                                    shapefile.rewind([(11,11),(11,19),(19,19),(19,11),(11,11)]), # exterior
-                                    shapefile.rewind([(12,12),(14,12),(14,14),(12,14),(12,12)]), # hole 1
-                                    shapefile.rewind([(15,15),(17,15),(17,17),(15,17),(15,15)]), # hole 2
+                                    [(11,11),(11,19),(19,19),(19,11),(11,11)], # exterior
+                                    [(12,12),(14,12),(14,14),(12,14),(12,12)], # hole 1
+                                    [(15,15),(17,15),(17,17),(15,17),(15,15)], # hole 2
                                 ],
-                                [ # poly 3 (orphaned hole) 
-                                  # Note: due to the hole-to-exterior conversion, should return the same ring orientation
+                                [ # poly 3 (orphaned hole)
                                     [(95,95),(97,95),(97,97),(95,97),(95,95)], # exterior
                                 ],
                                 ]}
@@ -184,11 +190,9 @@ geo_interface_tests = [ (shapefile.POINT, # point
                             [0,5],
                             {'type':'MultiPolygon','coordinates':[
                                 [ # poly 1
-                                  # Note: due to the hole-to-exterior conversion, should return the same ring orientation
                                     [(1,1),(9,1),(9,9),(1,9),(1,1)],
                                 ],
                                 [ # poly 2
-                                  # Note: due to the hole-to-exterior conversion, should return the same ring orientation
                                     [(11,11),(19,11),(19,19),(11,19),(11,11)],
                                 ],
                                 ]}
@@ -260,6 +264,7 @@ def test_reader_context_manager():
     assert sf.shx.closed is True
 
 
+@pytest.mark.network
 def test_reader_url():
     """
     Assert that Reader can open shapefiles from a url.
@@ -402,6 +407,65 @@ def test_reader_shapefile_extension_ignored():
     assert not os.path.exists(filename)
 
 
+def test_reader_pathlike():
+    """
+    Assert that path-like objects can be read.
+    """
+    base = Path("shapefiles")
+    with shapefile.Reader(base / "blockgroups") as sf:
+        assert len(sf) == 663
+
+
+def test_reader_dbf_only():
+    """
+    Assert that specifying just the
+    dbf argument to the shapefile reader
+    reads just the dbf file.
+    """
+    with shapefile.Reader(dbf="shapefiles/blockgroups.dbf") as sf:
+        assert len(sf) == 663
+        record = sf.record(3)
+        assert record[1:3] == ['060750601001', 4715]
+
+
+def test_reader_shp_shx_only():
+    """
+    Assert that specifying just the
+    shp and shx argument to the shapefile reader
+    reads just the shp and shx file.
+    """
+    with shapefile.Reader(shp="shapefiles/blockgroups.shp", shx="shapefiles/blockgroups.shx") as sf:
+        assert len(sf) == 663
+        shape = sf.shape(3)
+        assert len(shape.points) is 173
+
+    
+def test_reader_shp_dbf_only():
+    """
+    Assert that specifying just the
+    shp and shx argument to the shapefile reader
+    reads just the shp and dbf file.
+    """
+    with shapefile.Reader(shp="shapefiles/blockgroups.shp", dbf="shapefiles/blockgroups.dbf") as sf:
+        assert len(sf) == 663
+        shape = sf.shape(3)
+        assert len(shape.points) is 173
+        record = sf.record(3)
+        assert record[1:3] == ['060750601001', 4715]
+
+
+def test_reader_shp_only():
+    """
+    Assert that specifying just the
+    shp argument to the shapefile reader
+    reads just the shp file (shx optional).
+    """
+    with shapefile.Reader(shp="shapefiles/blockgroups.shp") as sf:
+        assert len(sf) == 663
+        shape = sf.shape(3)
+        assert len(shape.points) is 173
+
+
 def test_reader_filelike_dbf_only():
     """
     Assert that specifying just the
@@ -426,7 +490,21 @@ def test_reader_filelike_shp_shx_only():
         assert len(shape.points) is 173
 
 
-def test_reader_filelike_shx_optional():
+def test_reader_filelike_shp_dbf_only():
+    """
+    Assert that specifying just the
+    shp and shx argument to the shapefile reader
+    reads just the shp and dbf file.
+    """
+    with shapefile.Reader(shp=open("shapefiles/blockgroups.shp", "rb"), dbf=open("shapefiles/blockgroups.dbf", "rb")) as sf:
+        assert len(sf) == 663
+        shape = sf.shape(3)
+        assert len(shape.points) is 173
+        record = sf.record(3)
+        assert record[1:3] == ['060750601001', 4715]
+
+
+def test_reader_filelike_shp_only():
     """
     Assert that specifying just the
     shp argument to the shapefile reader
@@ -602,6 +680,194 @@ def test_shape_oid():
             assert shaperec.shape.oid == i
 
 
+def test_shape_oid_no_shx():
+    """
+    Assert that the shape's oid attribute returns
+    its index in the shapefile, when shx file is missing.
+    """
+    basename = "shapefiles/blockgroups"
+    shp = open(basename + ".shp", 'rb')
+    dbf = open(basename + ".dbf", 'rb')
+    with shapefile.Reader(shp=shp, dbf=dbf) as sf, \
+        shapefile.Reader(basename) as sf_expected:
+        for i in range(len(sf)):
+            shape = sf.shape(i)
+            assert shape.oid == i
+            shape_expected = sf_expected.shape(i)
+            assert shape.__geo_interface__ == shape_expected.__geo_interface__
+
+        for i,shape in enumerate(sf.shapes()):
+            assert shape.oid == i
+            shape_expected = sf_expected.shape(i)
+            assert shape.__geo_interface__ == shape_expected.__geo_interface__
+
+        for i,shape in enumerate(sf.iterShapes()):
+            assert shape.oid == i
+            shape_expected = sf_expected.shape(i)
+            assert shape.__geo_interface__ == shape_expected.__geo_interface__
+
+        for i,shaperec in enumerate(sf.iterShapeRecords()):
+            assert shaperec.shape.oid == i
+            shape_expected = sf_expected.shape(i)
+            assert shaperec.shape.__geo_interface__ == shape_expected.__geo_interface__
+
+
+def test_reader_offsets():
+    """
+    Assert that reader will not read the shx offsets unless necessary, 
+    i.e. requesting a shape index. 
+    """
+    basename = "shapefiles/blockgroups"
+    with shapefile.Reader(basename) as sf:
+        # shx offsets should not be read during loading
+        assert not sf._offsets
+        # reading a shape index should trigger reading offsets from shx file
+        shape = sf.shape(3)
+        assert len(sf._offsets) == len(sf.shapes())
+
+
+def test_reader_offsets_no_shx():
+    """
+    Assert that reading a shapefile without a shx file will not build 
+    the offsets unless necessary, i.e. reading all the shapes. 
+    """
+    basename = "shapefiles/blockgroups"
+    shp = open(basename + ".shp", 'rb')
+    dbf = open(basename + ".dbf", 'rb')
+    with shapefile.Reader(shp=shp, dbf=dbf) as sf:
+        # offsets should not be built during loading
+        assert not sf._offsets
+        # reading a shape index should iterate to the shape
+        # but the list of offsets should remain empty
+        shape = sf.shape(3)
+        assert not sf._offsets
+        # reading all the shapes should build the list of offsets
+        shapes = sf.shapes()
+        assert len(sf._offsets) == len(shapes)
+
+
+
+def test_reader_numshapes():
+    """
+    Assert that reader reads the numShapes attribute from the
+    shx file header during loading. 
+    """
+    basename = "shapefiles/blockgroups"
+    with shapefile.Reader(basename) as sf:
+        # numShapes should be set during loading
+        assert sf.numShapes != None
+        # numShapes should equal the number of shapes
+        assert sf.numShapes == len(sf.shapes())
+
+
+def test_reader_numshapes_no_shx():
+    """
+    Assert that reading a shapefile without a shx file will have
+    an unknown value for the numShapes attribute (None), and that
+    reading all the shapes will set the numShapes attribute.
+    """
+    basename = "shapefiles/blockgroups"
+    shp = open(basename + ".shp", 'rb')
+    dbf = open(basename + ".dbf", 'rb')
+    with shapefile.Reader(shp=shp, dbf=dbf) as sf:
+        # numShapes should be unknown due to missing shx file
+        assert sf.numShapes == None
+        # numShapes should be set after reading all the shapes
+        shapes = sf.shapes()
+        assert sf.numShapes == len(shapes)
+
+
+def test_reader_len():
+    """
+    Assert that calling len() on reader is equal to length of 
+    all shapes and records. 
+    """
+    basename = "shapefiles/blockgroups"
+    with shapefile.Reader(basename) as sf:
+        assert len(sf) == len(sf.records()) == len(sf.shapes())
+
+
+def test_reader_len_not_loaded():
+    """
+    Assert that calling len() on reader that hasn't loaded a shapefile
+    yet is equal to 0. 
+    """
+    with shapefile.Reader() as sf:
+        assert len(sf) == 0
+
+
+def test_reader_len_dbf_only():
+    """
+    Assert that calling len() on reader when reading a dbf file only,
+    is equal to length of all records. 
+    """
+    basename = "shapefiles/blockgroups"
+    dbf = open(basename + ".dbf", 'rb')
+    with shapefile.Reader(dbf=dbf) as sf:
+        assert len(sf) == len(sf.records())
+
+
+def test_reader_len_no_dbf():
+    """
+    Assert that calling len() on reader when dbf file is missing,
+    is equal to length of all shapes. 
+    """
+    basename = "shapefiles/blockgroups"
+    shp = open(basename + ".shp", 'rb')
+    shx = open(basename + ".shx", 'rb')
+    with shapefile.Reader(shp=shp, shx=shx) as sf:
+        assert len(sf) == len(sf.shapes())
+
+
+def test_reader_len_no_dbf_shx():
+    """
+    Assert that calling len() on reader when dbf and shx file is missing,
+    is equal to length of all shapes. 
+    """
+    basename = "shapefiles/blockgroups"
+    shp = open(basename + ".shp", 'rb')
+    with shapefile.Reader(shp=shp) as sf:
+        assert len(sf) == len(sf.shapes())
+
+
+def test_reader_corrupt_files():
+    """
+    Assert that reader is able to handle corrupt files by
+    strictly going off the header information. 
+    """
+    basename = "shapefiles/test/corrupt_too_long"
+
+    # write a shapefile with junk byte data at end of files
+    with shapefile.Writer(basename) as w:
+        w.field("test", "C", 50)
+        # add 10 line geoms
+        for _ in range(10):
+            w.record("value")
+            w.line([[(1,1),(1,2),(2,2)]])
+        # add junk byte data to end of dbf and shp files
+        w.dbf.write(b'12345')
+        w.shp.write(b'12345')
+
+    # read the corrupt shapefile and assert that it reads correctly
+    with shapefile.Reader(basename) as sf:
+        # assert correct shapefile length metadata
+        assert len(sf) == sf.numRecords == sf.numShapes == 10
+        # assert that records are read without error
+        assert len(sf.records()) == 10
+        # assert that didn't read the extra junk data
+        stopped = sf.dbf.tell()
+        sf.dbf.seek(0, 2)
+        end = sf.dbf.tell()
+        assert (end - stopped) == 5
+        # assert that shapes are read without error
+        assert len(sf.shapes()) == 10
+        # assert that didn't read the extra junk data
+        stopped = sf.shp.tell()
+        sf.shp.seek(0, 2)
+        end = sf.shp.tell()
+        assert (end - stopped) == 5
+
+
 def test_bboxfilter_shape():
     """
     Assert that applying the bbox filter to shape() correctly ignores the shape
@@ -817,38 +1083,93 @@ def test_write_shp_only(tmpdir):
     shp argument to the shapefile writer
     creates just a shp file.
     """
-    filename = tmpdir.join("test.shp").strpath
-    with shapefile.Writer(shp=filename) as writer:
-        pass
+    filename = tmpdir.join("test").strpath
+    with shapefile.Writer(shp=open(filename+'.shp','wb')) as writer:
+        writer.point(1, 1)
+    assert writer.shp and not writer.shx and not writer.dbf
+    assert writer.shpNum == 1
+    assert len(writer) == 1
+    assert writer.shp.closed == True
 
     # assert test.shp exists
-    assert os.path.exists(filename)
+    assert os.path.exists(filename+'.shp')
+    
+    # test that can read shapes
+    with shapefile.Reader(shp=open(filename+'.shp','rb')) as reader:
+        assert reader.shp and not reader.shx and not reader.dbf
+        assert (reader.numRecords, reader.numShapes) == (None, None) # numShapes is unknown in the absence of shx file
+        assert len(reader.shapes()) == 1
 
     # assert test.shx does not exist
-    assert not os.path.exists(tmpdir.join("test.shx").strpath)
+    assert not os.path.exists(filename+'.shx')
 
     # assert test.dbf does not exist
-    assert not os.path.exists(tmpdir.join("test.dbf").strpath)
+    assert not os.path.exists(filename+'.dbf')
 
 
-def test_write_shx_only(tmpdir):
+def test_write_shp_shx_only(tmpdir):
     """
-    Assert that specifying just the
+    Assert that specifying just the shp and
     shx argument to the shapefile writer
-    creates just a shx file.
+    creates just a shp and shx file.
     """
-    filename = tmpdir.join("test.shx").strpath
-    with shapefile.Writer(shx=filename) as writer:
-        pass
+    filename = tmpdir.join("test").strpath
+    with shapefile.Writer(shp=open(filename+'.shp','wb'), shx=open(filename+'.shx','wb')) as writer:
+        writer.point(1, 1)
+    assert writer.shp and writer.shx and not writer.dbf
+    assert writer.shpNum == 1
+    assert len(writer) == 1
+    assert writer.shp.closed == writer.shx.closed == True
+
+    # assert test.shp exists
+    assert os.path.exists(filename+'.shp')
 
     # assert test.shx exists
-    assert os.path.exists(filename)
+    assert os.path.exists(filename+'.shx')
 
-    # assert test.shp does not exist
-    assert not os.path.exists(tmpdir.join("test.shp").strpath)
+    # test that can read shapes and offsets
+    with shapefile.Reader(shp=open(filename+'.shp','rb'), shx=open(filename+'.shx','rb')) as reader:
+        assert reader.shp and reader.shx and not reader.dbf
+        assert (reader.numRecords, reader.numShapes) == (None, 1)
+        reader.shape(0) # trigger reading of shx offsets
+        assert len(reader._offsets) == 1
+        assert len(reader.shapes()) == 1
 
     # assert test.dbf does not exist
-    assert not os.path.exists(tmpdir.join("test.dbf").strpath)
+    assert not os.path.exists(filename+'.dbf')
+
+
+def test_write_shp_dbf_only(tmpdir):
+    """
+    Assert that specifying just the
+    shp and dbf argument to the shapefile writer
+    creates just a shp and dbf file.
+    """
+    filename = tmpdir.join("test").strpath
+    with shapefile.Writer(shp=open(filename+'.shp','wb'), dbf=open(filename+'.dbf','wb')) as writer:
+        writer.field('field1', 'C') # required to create a valid dbf file
+        writer.record('value')
+        writer.point(1, 1)
+    assert writer.shp and not writer.shx and writer.dbf
+    assert writer.shpNum == writer.recNum == 1
+    assert len(writer) == 1
+    assert writer.shp.closed == writer.dbf.closed == True
+
+    # assert test.shp exists
+    assert os.path.exists(filename+'.shp')
+
+    # assert test.dbf exists
+    assert os.path.exists(filename+'.dbf')
+    
+    # test that can read records and shapes
+    with shapefile.Reader(shp=open(filename+'.shp','rb'), dbf=open(filename+'.dbf','rb')) as reader:
+        assert reader.shp and not reader.shx and reader.dbf
+        assert (reader.numRecords, reader.numShapes) == (1, None) # numShapes is unknown in the absence of shx file
+        assert len(reader.records()) == 1
+        assert len(reader.shapes()) == 1
+
+    # assert test.shx does not exist
+    assert not os.path.exists(filename+'.shx')
 
 
 def test_write_dbf_only(tmpdir):
@@ -857,18 +1178,29 @@ def test_write_dbf_only(tmpdir):
     dbf argument to the shapefile writer
     creates just a dbf file.
     """
-    filename = tmpdir.join("test.dbf").strpath
-    with shapefile.Writer(dbf=filename) as writer:
+    filename = tmpdir.join("test").strpath
+    with shapefile.Writer(dbf=open(filename+'.dbf','wb')) as writer:
         writer.field('field1', 'C') # required to create a valid dbf file
+        writer.record('value')
+    assert not writer.shp and not writer.shx and writer.dbf
+    assert writer.recNum == 1
+    assert len(writer) == 1
+    assert writer.dbf.closed == True
 
     # assert test.dbf exists
-    assert os.path.exists(filename)
+    assert os.path.exists(filename+'.dbf')
+
+    # test that can read records
+    with shapefile.Reader(dbf=open(filename+'.dbf','rb')) as reader:
+        assert not writer.shp and not writer.shx and writer.dbf
+        assert (reader.numRecords, reader.numShapes) == (1, None)
+        assert len(reader.records()) == 1
 
     # assert test.shp does not exist
-    assert not os.path.exists(tmpdir.join("test.shp").strpath)
+    assert not os.path.exists(filename+'.shp')
 
     # assert test.shx does not exist
-    assert not os.path.exists(tmpdir.join("test.shx").strpath)
+    assert not os.path.exists(filename+'.shx')
 
 
 def test_write_default_shp_shx_dbf(tmpdir):
@@ -887,6 +1219,20 @@ def test_write_default_shp_shx_dbf(tmpdir):
     assert os.path.exists(filename + ".dbf")
 
 
+def test_write_pathlike(tmpdir):
+    """
+    Assert that path-like objects can be written.
+    Similar to test_write_default_shp_shx_dbf.
+    """
+    filename = tmpdir.join("test")
+    assert not isinstance(filename, str)
+    with shapefile.Writer(filename) as writer:
+        writer.field('field1', 'C')
+    assert (filename + ".shp").ensure()
+    assert (filename + ".shx").ensure()
+    assert (filename + ".dbf").ensure()
+
+
 def test_write_shapefile_extension_ignored(tmpdir):
     """
     Assert that the filename's extension is
@@ -917,10 +1263,10 @@ def test_write_record(tmpdir):
     with shapefile.Writer(filename) as writer:
         writer.autoBalance = True
 
-        writer.field('one', 'C') # many under length limit
-        writer.field('two', 'C') # 1 under length limit
-        writer.field('three', 'C') # at length limit
-        writer.field('four', 'C') # 1 over length limit
+        writer.field('one', 'C') 
+        writer.field('two', 'C') 
+        writer.field('three', 'C') 
+        writer.field('four', 'C') 
         
         values = ['one','two','three','four']
         writer.record(*values)
@@ -944,10 +1290,10 @@ def test_write_partial_record(tmpdir):
     with shapefile.Writer(filename) as writer:
         writer.autoBalance = True
         
-        writer.field('one', 'C') # many under length limit
-        writer.field('two', 'C') # 1 under length limit
-        writer.field('three', 'C') # at length limit
-        writer.field('four', 'C') # 1 over length limit
+        writer.field('one', 'C') 
+        writer.field('two', 'C') 
+        writer.field('three', 'C') 
+        writer.field('four', 'C') 
         
         values = ['one','two']
         writer.record(*values)
@@ -1004,4 +1350,11 @@ def test_write_empty_shapefile(tmpdir, shape_type):
         w.field('field1', 'C') # required to create a valid dbf file
 
     with shapefile.Reader(filename) as r:
+        # test correct shape type
         assert r.shapeType == shape_type
+        # test length 0
+        assert len(r) == r.numRecords == r.numShapes == 0
+        # test records are empty
+        assert len(r.records()) == 0
+        # test shapes are empty
+        assert len(r.shapes()) == 0
